#include <string>
#include <fstream>
#include <new>
#include <iostream>

#include "MyTmpltLib.h"

using namespace std;

#include "Course.cpp"
#include "Student.cpp"
#include "StudentInCourse.cpp"
#include "StudentInCourseCollection.h"
#include "CourseCollection.h"
#include "SIndexEntry.h"
#include "StudentCollection.h"
#include "CourseCollection.cpp"
#include "StudentCollection.cpp"
#include "StudentInCourseCollection.cpp"

struct ProgXptn
{
   enum { allocFailed, cannotOpen, cannotRead };
   char functionName[100];
   int  errorCode;
   char errStrVal[100];
   ProgXptn( const char* fn, int ec, const char* sv="" )
      :  errorCode( ec )
   {  strncpy( functionName, fn, 99 );  functionName[99] = '\0';
      strncpy( errStrVal, sv, 99 );  errStrVal[99] = '\0';  }
}; // ProgXptn

void loadCourses( string flNm, CourseCollection& allCourses );
void readStudentData( string flNm,
                      StudentCollection& allStudents,
                      StudentInCourseCollection& allEnrollments,
                      CourseCollection& allCourses,
                      ofstream& log );
void saveCollections( CourseCollection& allCourses,
                      StudentCollection& allStudents,
                      SIndexEntry* sIndex,
                      StudentInCourseCollection& allEnrollments );

int main()
{
   try
   {
      CourseCollection allCourses;
      StudentCollection allStudents;
      StudentInCourseCollection allEnrollments;
      
      allEnrollments.setPAllStudents( &allStudents );
      allEnrollments.setPAllCourses( &allCourses );
      
//      allCourses.setPAllEnrollments( &allEnrollments );
      allStudents.setPAllEnrollments( &allEnrollments );

      loadCourses( "gCourses.dta", allCourses );
      
      ofstream log( "log.txt" );
      if ( log.fail() )
         throw ProgXptn( "main", ProgXptn::cannotOpen, "log.txt" );

      readStudentData( "enrllmnt.txt", allStudents, allEnrollments,
                       allCourses, log );
                       
      unsigned int maxWH( 30 ); //   
                                // : 30 /
      allStudents.checkWH( log, maxWH );

      log.close();

      SIndexEntry* sIndex;
      try
      {  sIndex = new SIndexEntry[allStudents.getNOfStudents()];  }
      catch( bad_alloc )
      {  throw ProgXptn( "main", ProgXptn::allocFailed );  }
      saveCollections( allCourses, allStudents, sIndex,
                       allEnrollments );

      delete[] sIndex;
   } // try
   catch( ProgXptn& x )
   {
      switch ( x.errorCode )
      {
         case ProgXptn::allocFailed:
            cout << "no dynamic memory in " << x.functionName << endl;
            break;
         case ProgXptn::cannotOpen:
            cout << "cannot open file " << x.errStrVal << " in "
                 << x.functionName << endl;
            break;
         default:
            cout << "unexpected ProgXptn from " << x.functionName << endl;
      } // switch
   } // catch( ProgXptn
   catch( MyTmpltLibXptn& x )
   {
      switch ( x.errCode )
      {
         case MyTmpltLibXptn::noArray:
            cout << x.funcName << " called with NULL array pointer"
                 << endl;
            break;
         case MyTmpltLibXptn::paramError:
            cout << x.funcName << ": param error ni(" << x.errIntVal1
                 << ") and nf(" << x.errIntVal2 << ")" << endl;
            break;
         case MyTmpltLibXptn::allocFailed:
            cout << "no dynamic memory in " << x.funcName << endl;
            break;
         case MyTmpltLibXptn::cannotCreate:
            cout << "cannot create file " << x.errStrVal << " in "
                 << x.funcName << endl;
            break;
         case MyTmpltLibXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.funcName << endl;
            break;
         default:
            cout << "unexpected MyTmpltLibXptn from "
                 << x.funcName << endl;
      } // switch
   } // catch( MyTmpltLibXptn
   catch( CourseXptn& x )
   {
      switch ( x.errorCode )
      {
         case CourseXptn::keyLen:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " illegal key length (" << x.errStrVal << ")" << endl;
            break;
         case CourseXptn::rangeError:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " range error (" << x.errIntVal << ")" << endl;
            break;
         case CourseXptn::noSuchSector:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " no such sector (" << char(x.errIntVal) << ")" << endl;
            break;
         case CourseXptn::noSuchCateg:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " no such category (" << x.errStrVal << ")" << endl;
            break;
         case CourseXptn::autoRef:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " prerequisite of itself" << endl;
            break;
         case CourseXptn::noStudent:
            cout << "course " << x.objKey.s << " in " << x.methodName
                 << " no student to delete" << endl;
            break;
         case CourseXptn::fileNotOpen:
            cout << "course " << x.objKey.s << ": file " << x.errStrVal
                 << " not open in " << x.methodName << endl;
            break;
         case CourseXptn::cannotWrite:
            cout << "cannot write " << x.objKey.s << " to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case CourseXptn::cannotRead:
            cout << "cannot read " << x.objKey.s << " from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected CourseXptn from " << x.methodName
                 << " of " << x.objKey.s << endl;
      } // switch
   } // catch( CourseXptn
   catch( StudentXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentXptn::allocFailed:
            cout << "Student " << x.objKey
                 << ": no dynamic memory in " << x.methodName << endl;
            break;
         case StudentXptn::negIdNum:
            cout << "Student " << x.objKey << ": illegal id num ("
                 << x.errIntVal << ") in " << x.methodName << endl;
            break;
         case StudentXptn::incomplete:
            cout << "incomplete data for Student " << x.objKey
                 << " in " << x.methodName << endl;
            break;
         case StudentXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case StudentXptn::cannotRead:
            cout << "cannot read from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentXptn from " << x.methodName << endl;
      } // switch
   } // catch( StudentXptn
   catch( StudentInCourseXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentInCourseXptn::negIdNum:
            cout << x.methodName << ": illegal Student id num ("
                 << x.objKey.sIdNum << ") in course "
                 << x.objKey.CCode.s <<endl;
            break;
         case StudentInCourseXptn::keyLen:
            cout << x.methodName << ": illegal course key length ("
                 << x.errStrVal << ") from student "
                 << x.objKey.sIdNum << endl;
            break;
         case StudentInCourseXptn::rangeError:
            cout << x.methodName << ": range error (" << x.errFltVal
                  << ") in course for student " << x.objKey.sIdNum
                  << " in course " << x.objKey.CCode.s <<endl;
            break;
         case StudentInCourseXptn::unknownCCode:
            cout << "unknown Course Code (" << x.errStrVal << ") in "
                 << x.methodName << endl;
            break;
         case StudentInCourseXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentInCourseXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentInCourseXptn for Student "
                 << x.objKey.sIdNum << " in course " << x.objKey.CCode.s
                 << " from " << x.methodName << endl;
      } // switch
   } // catch( StudentInCourseXptn
   catch( CourseCollectionXptn& x )
   {
      switch ( x.errorCode )
      {
         case CourseCollectionXptn::allocFailed:
            cout << "no dynamic memory in " << x.methodName << endl;
            break;
         case CourseCollectionXptn::notFound:
            cout << "no Course with code " << x.errStrVal
                 << " in " << x.methodName << endl;
            break;
         case CourseCollectionXptn::entity:
            cout << "Course with null key (code) " << " in " << x.methodName << endl;
            break;
         case CourseCollectionXptn::prereqRef:
            cout << "Prerequisite with code " << x.errStrVal << " not in table in "
                  << x.methodName << endl;
            break;
         case CourseCollectionXptn::noEnroll:
            cout << "Enrollments table not accessible in " << x.methodName << endl;
            break;
         case CourseCollectionXptn::enrollRef:
            cout << "Cannot delete course " << x.errStrVal << " in " << x.methodName
                 << ". Student " << x.errIntVal << " enrolled" << endl;
            break;
         case CourseCollectionXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case CourseCollectionXptn::cannotWrite:
            cout << "cannot write " << " to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case CourseCollectionXptn::cannotRead:
            cout << "cannot read " << " from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected CourseCollectionXptn from "
                 << x.methodName << endl;
      } // switch
   } // catch( CourseCollectionXptn
   catch( StudentCollectionXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentCollectionXptn::allocFailed:
            cout << "no dynamic memory in " << x.methodName << endl;
            break;
         case StudentCollectionXptn::notFound:
            cout << "no Student with Id Number " << x.errIntVal
                 << " in " << x.methodName << endl;
            break;
         case StudentCollectionXptn::noEnroll:
            cout << "Enrollments table not accessible in " << x.methodName << endl;
            break;
         case StudentCollectionXptn::enrollRef:
            cout << "Cannot delete Student " << x.errIntVal << " in " << x.methodName
                 << " enrolled in course " << x.errStrVal << endl;
            break;
         case StudentCollectionXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentCollectionXptn::cannotWrite:
            cout << "cannot write " << " to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case StudentCollectionXptn::cannotRead:
            cout << "cannot read " << " from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentCollectionXptn from "
                 << x.methodName << endl;
      } // switch
   } // catch( StudentCollectionXptn
   catch( StudentInCourseCollectionXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentInCourseCollectionXptn::allocFailed:
            cout << "no dynamic memory in " << x.methodName << endl;
            break;
         case StudentInCourseCollectionXptn::notFound:
            cout << x.methodName << "cannot find enrollment of "
                 << x.errSICVal.sIdNum
                 << " in " << x.errSICVal.CCode.s << endl;
            break;
         case StudentInCourseCollectionXptn::noCrs:
            cout << x.methodName << "cannot find course table"
                 << endl;
            break;
         case StudentInCourseCollectionXptn::unknownCrs:
            cout << "unknown Course Code (" << x.errStrVal << ") in "
                 << x.methodName << endl;
            break;
         case StudentInCourseCollectionXptn::noStdnt:
            cout << x.methodName << "cannot find student table"
                 << endl;
            break;
         case StudentInCourseCollectionXptn::unknownStdn:
            cout << "no student with idNum " << x.errIntVal << " in "
                 << x.methodName << endl;
            break;
         case StudentInCourseCollectionXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentInCourseCollectionXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case StudentInCourseCollectionXptn::cannotRead:
            cout << "cannot read from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentInCourseCollectionXptn from "
                 << x.methodName << endl;
      } // switch
   } // catch( SStudentInCourseCollectionXptn
   catch( ... )
   {
      cout << "unexpected exceptioon" << endl;
   }
system("pause");
} // main

void loadSylCourse( ifstream& bin, Course& oneCourse )
{
   bool         bBuf;
   unsigned int iBuf;
   char         cBuf[Course::cTitleSz];

   bin.read( cBuf, Course::cCodeSz );         // ??d???? a??at??
   if ( !bin.eof() )
   {
                                        oneCourse.setCode( cBuf );
      bin.read( cBuf, Course::cTitleSz );      // t?t??? a??at??
                                       oneCourse.setTitle( cBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );          // t?p??? e?????
                                        oneCourse.setFSem( iBuf );
      bin.read( reinterpret_cast<char*>(&bBuf),
                sizeof(bool) );          // ?p???e?t??? ? ep??????
                                     oneCourse.setCompuls( bBuf );
      bin.read( cBuf, sizeof(char) );                    // t??a?
                                   oneCourse.setSector( cBuf[0] );
      bin.read( cBuf, Course::cCategSz );             // ?at?????a
                                       oneCourse.setCateg( cBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );       // ??e? a?? ed??da
                                          oneCourse.setWH( iBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );      // d?da?t???? ???de?
                                       oneCourse.setUnits( iBuf );
      bin.read( cBuf, Course::cCodeSz );         // p??apa?t??e??
                                      oneCourse.setPrereq( cBuf );
      if ( bin.fail() )
         throw ProgXptn( "loadSylCourse", ProgXptn::cannotRead );
   }
} // loadSylCourse

void loadCourses( string flNm, CourseCollection& allCourses )
{
//   
   ifstream bin( flNm.c_str(), ios_base::binary );
   if ( bin.fail() )
   {
      throw ProgXptn( "loadCourses", ProgXptn::cannotOpen, flNm.c_str() );
   }
//  
   Course oneCourse;
   loadSylCourse( bin, oneCourse );
   while ( !bin.eof() )
   {
      oneCourse.clearStudents();
      allCourses.add1Course( oneCourse );
      loadSylCourse( bin, oneCourse );
   } // while
   bin.close();
} // loadCourses

void ignoreStudentData( istream& tin )
{
   string str1;
   getline( tin, str1, '\n' );
   int noc( atoi(str1.c_str()) );
   for ( int k(0); k <= noc && !tin.eof(); ++k )
      getline( tin, str1, '\n' );
} // ignoreStudentData

void readAStudent( istream& tin,
                   StudentCollection& allStudents,
                   StudentInCourseCollection& allEnrollments,
                   CourseCollection& allCourses,
                   ofstream& log )
{
   Student      aStudent;
   unsigned int sIdNum;
       
//     
   aStudent.readPartFromText( tin );
   if ( !tin.eof() )
   {
      sIdNum = aStudent.getIdNum();
      if ( allStudents.find1Student(aStudent.getIdNum()) )
      { //    
      //   
         ignoreStudentData( tin );
         log << " multiple entry for student with id num "
             << sIdNum << endl;
      }
      else
      {
      //      
         allStudents.add1Student( aStudent );
         string str1;
         getline( tin, str1, '\n' );
         if ( !tin.eof() )
         {
            int noc( atoi(str1.c_str()) );
            for ( int k(0); k < noc && !tin.eof(); ++k )
            {
               getline( tin, str1, '\n' );
               try
               {
                  StudentInCourse aStdnInCrs( sIdNum, str1 );
                  allEnrollments.add1StudentInCourse( aStdnInCrs );
               }
               catch( StudentInCourseXptn& x )
               {
                  if ( x.errorCode ==
                                     StudentInCourseXptn::keyLen )
                     log << " student with id num "
                         << aStudent.getIdNum()
                         << " asking course " << str1 << endl;
                  else
                     throw;
               } // catch( StudentInCourseXptn
               catch( StudentInCourseCollectionXptn& x )
               {
                  if ( x.errorCode ==
                       StudentInCourseCollectionXptn::unknownCrs )
                     log << " student with id num "
                         << aStudent.getIdNum()
                         << " asking course " << str1 << endl;
                  else
                     throw;
               } // catch( StudentInCourseCollectionXptn
            } // for
         } // if ( !tin.eof() )
         if ( !tin.eof() ) getline( tin, str1, '\n' ); // blank line
      } // if ( allStudents.find1Student(aStudent.getIdNum()) )
   } // if ( !tin.eof() )
} // readAStudent

void readStudentData( string flNm,
                      StudentCollection& allStudents,
                      StudentInCourseCollection& allEnrollments,
                      CourseCollection& allCourses,
                      ofstream& log )
{
   ifstream tin( flNm.c_str() ); //"enrllmnt.txt" );
   if ( tin.fail() )
      throw ProgXptn( "readStudentData", ProgXptn::cannotOpen, flNm.c_str() );
   do {
      readAStudent( tin, allStudents, allEnrollments, allCourses,
                    log );
   } while( !tin.eof() );
   tin.close();
} // readStudentData

void saveCollections( CourseCollection& allCourses,
                      StudentCollection& allStudents,
                      SIndexEntry* index,
                      StudentInCourseCollection& allEnrollments )
{
   ofstream bout( "Courses.dta", ios_base::binary );
   allCourses.save( bout );
   bout.close();
   bout.open( "students.dta", ios_base::binary );
   allStudents.save( bout, index );
   bout.close();
   bout.open( "students.ndx", ios_base::binary );
   for ( int k(0); k < allStudents.getNOfStudents(); ++k )
      bout.write( reinterpret_cast<const char*>(&index[k]),
                  sizeof(SIndexEntry) );
   bout.close();
   bout.open( "enrllmnt.dta", ios_base::binary );
   allEnrollments.save( bout );
   bout.close();
} // saveCollections

